/**
 * @file LispParser.hpp
 * Defines class LispParser for parsing Lisp files.
 *
 * @since 26/08/2009 Redmond
 */

#ifndef __LispParser__
#define __LispParser__

#include <string>

#include "Forwards.hpp"
#include "Token.hpp"

#include "Lib/Exception.hpp"
#include "Lib/List.hpp"
#include "Lib/Portability.hpp"
#include "Lib/Stack.hpp"

namespace Shell {

using namespace std;
using namespace Lib;

class LispLexer;

/**
 * Class LispParser, implements a Lisp Parser.
 * @since 26/08/2009 Redmond
 */
class LispParser
{
public:
  /** Tags of Lisp expressions */
  enum Tag {
    /** atom */
    ATOM = 0,
    /** list */
    LIST = 1
  };

  /** expressions */
  struct Expression {
    /** type of the expression */
    Tag tag;
    /** the value (for atoms and numbers) */
    string str;
    /** list of expressions */
    List<Expression*>* list;
    /** build a list expressions with the list initially empty */
    explicit Expression(Tag t)
      : tag(t),
	str("?"),
	list(0)
    {}
    /** build a string-values expression */
    Expression(Tag t,string s)
      : tag(t),
	str(s),
	list(0)
    {}
    string toString(bool outerParentheses=true) const;

    bool isList() const { return tag==LIST; }
    bool isAtom() const { return tag==ATOM; }

    bool get2Args(string functionName, Expression*& arg1, Expression*& arg2);
    bool get1Arg(string functionName, Expression*& arg);
    bool getPair(Expression*& el1, Expression*& el2);
    bool getSingleton(Expression*& el);
  };

  typedef Lib::List<Expression*> List;

  explicit LispParser(LispLexer& lexer);
  Expression* parse();
  void parse(List**);

  /**
   * Class Exception. Implements parser exceptions.
   * @since 17/07/2004 Helsinki airport
   */
  class Exception 
    : public Lib::Exception
  {
  public:                                
    Exception (string message,const Token&);
    void cry (ostream&);
    ~Exception () {}
  protected:
    string _message;
  }; // Exception

private:
  /** lexer supplying tokens */
  LispLexer& _lexer;
  /** balance of parenthesis */
  int _balance;
}; // class LispParser

typedef LispParser::Expression LExpr;
typedef List<LExpr*> LExprList;


class LispListReader {
public:
  explicit LispListReader(LExpr* e)
  {
    CALL("LispListReader::LispListReader(LExpr*)");
    if(!e->isList()) {
      lispError(e, "list expected");
    }
    it = LExprList::Iterator(e->list);
  }
  explicit LispListReader(LExprList* list) : it(list) {}

  void lispError(LExpr* expr, string reason="error") NO_RETURN;
  void lispCurrError(string reason="error") NO_RETURN;

  bool hasNext() { return it.hasNext(); }
  LExpr* peekAtNext();
  LExpr* readNext();
  LExpr* next() { return readNext(); }

  bool tryReadAtom(string& atom);
  string readAtom();

  bool tryReadListExpr(LExpr*& e);
  LExpr* readListExpr();

  bool tryReadList(LExprList*& list);
  LExprList* readList();

  bool tryAcceptAtom(string atom);
  void acceptAtom(string atom);
  void acceptAtom() { readAtom(); }

  bool tryAcceptList();
  void acceptList();

  void acceptEOL();

  bool lookAheadAtom(string atom);

  bool tryAcceptCurlyBrackets();
private:
  LExprList::Iterator it;
};

class LispListWriter
{
public:
  LispListWriter()
  {
#if VDEBUG
    _destroyed = false;
#endif
  }

#if VDEBUG
  ~LispListWriter()
  {
    _destroyed = true;
  }
#endif

  LispListWriter& operator<<(string s)
  {
    _elements.push(new LExpr(LispParser::ATOM, s));
    return *this;
  }

  LispListWriter& operator<<(LExpr* e)
  {
    _elements.push(e);
    return *this;
  }

  LispListWriter& operator<<(const LispListWriter& e)
  {
    _elements.push(e.get());
    return *this;
  }

  LispListWriter& append(LExprList* lst)
  {
    _elements.loadFromIterator(LExprList::Iterator(lst));
    return *this;
  }

  LExprList* getList() const
  {
    CALL("LispListWriter::getList");
    ASS(!_destroyed);

    LExprList* res = 0;
    LExprList::pushFromIterator(Stack<LExpr*>::TopFirstIterator(_elements), res);
    return res;
  }

  LExpr* get() const
  {
    CALL("LispListWriter::get");

    LExpr* res = new LExpr(LispParser::LIST);
    res->list = getList();
    return res;
  }

private:
#if VDEBUG
  bool _destroyed;
#endif
  Stack<LExpr*> _elements;
};

}

#endif

